-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local C = {}
C.__index = C

local version = 1

local function numberToCurrency(num) 
  local str1 = tostring(num)

  -- 小数点以下の抽出
  local str2 = ""
  local match_s = string.find(str1, ".", 1, true)
  if match_s ~= nil then
    str2 = string.sub(str1, match_s)
    str1 = string.sub(str1, 1, match_s - 1)
  end

  -- 整数部分を後ろから3桁ずつ取り出して","で繋ぐ
  local i = 1
  str1 = string.reverse(str1)
  while (i <= string.len(str1)) do
    str2 = "," .. string.reverse(string.sub(str1, i, i + 2)) .. str2
    i = i + 3
  end

  -- 先頭に余分な","がついているので2文字目以降を返す
  return string.sub(str2, 2)
end

function C:init()
  self.latestVersion = version
  self.fgVariables = self.missionTypeData or {}
  self.fgPath = '/gameplay/missionTypes/mathkuroMission02/thrillDrive.flow.json'
  -- self.additionalAttributes.vehicle = self.missionTypeData.presetVehicleActive and 'provided' or 'own'

  self.defaultAggregateValues = {
    all = {
      bestMedal = 'none',
      bestPlace = 100,
      bestTime = nil
    }
  }

  self.defaultLeaderboardKey = 'highscore'
  self.defaultProgressKey = self.missionTypeData.rivalCarGroup

  self.starLabels = {
    justFinish = 'missions.missions.stars.justFinish',
    finishBronze = 'missions.aiRace.stars.bronze',
    finishSilver = 'missions.aiRace.stars.silver',
    finishGold = 'missions.aiRace.stars.gold',
    justFinishPlayerVeh = 'missions.aiRace.stars.justFinishPlayerVeh',
    finishGoldPlayerVeh = 'missions.aiRace.stars.finishGoldPlayerVeh',
    finishGoalLaps = 'missions.aiRace.stars.finishGoalLaps',
    beatGoalTime = 'missions.aiRace.stars.beatGoalTime',
    cleanRace = 'missions.aiRace.stars.cleanRace',
    cleanGoldRace = 'missions.aiRace.stars.cleanGoldRace',
    dirtyRace = 'missions.aiRace.stars.dirtyRace',
    trafficRace = 'missions.aiRace.stars.trafficRace',
    perfectRace = 'Perfect Race. (Achieve 1st place with no reset)',
    totalDamage = 'Achieve damage over $' .. numberToCurrency(self.missionTypeData.totalDamage),
    bounty = 'Achieve bounty over ' .. numberToCurrency(self.missionTypeData.bounty)
  }

  self.sortedStarKeys = {'justFinish', 'finishBronze', 'finishSilver', 'finishGold', 'justFinishPlayerVeh', 'finishGoldPlayerVeh', 'finishGoalLaps', 'beatGoalTime', 'cleanRace', 'cleanGoldRace', 'dirtyRace', 'trafficRace', 'perfectRace', 'totalDamage', 'bounty'}

  self.defaultStarOutroTexts = {
    noStarUnlocked = 'missions.aiRace.general.pass',
    justFinish = 'missions.aiRace.general.complete',
    finishBronze = 'missions.aiRace.general.complete',
    finishSilver = 'missions.aiRace.general.complete',
    finishGold = 'missions.aiRace.general.complete'
  }

  self.autoAggregates = {
    {
      type = 'simpleMedal',
      attemptKey = 'medal', -- key in the attempt
      aggregateKey = 'bestMedal', -- key in the aggregate
      newBestKey = 'newBestMedal',
      default = 'none',
    },
    {
      type = 'simpleHighscore',
      attemptKey = 'place', -- key in the attempt
      aggregateKey = 'bestPlace', -- key in the aggregate
      sorting = 'ascending', -- keeping the lower score
      newBestKey = 'newBestPlace',
    },
    {
      type = 'simpleHighscore',
      attemptKey = 'time', -- key in the attempt
      aggregateKey = 'bestTime', -- key in the aggregate
      sorting = 'ascending', -- keeping the lower time
      newBestKey = 'newBestTime',
      leaderboardKey = 'highscore',
      newLeaderboardEntryKey = 'rankedOnLeaderboardIndex'
    }
  }

  self.autoUiAttemptProgress = {
    {
      type = 'simple',
      attemptKey = 'medal',
      columnLabel = 'bigMap.progressLabels.medal',
    },
    {
      type = 'simple',
      attemptKey = 'place',
      columnLabel = 'bigMap.progressLabels.place',
    },
    {
      type = 'simple',
      formatFunction = 'detailledTime',
      attemptKey = 'time',
      columnLabel = 'bigMap.progressLabels.time',
    },
  }

  self.nautoUiAggregateProgress = {
    {
      type = 'simple',
      aggregateKey = 'bestMedal',
      columnLabel = 'bigMap.progressLabels.bestMedal',
      newBestKey = 'newBestMedal',
    },
    {
      type = 'simple',
      aggregateKey = 'bestPlace',
      columnLabel = 'bigMap.progressLabels.bestPlace',
      newBestKey = 'newBestPlace',
    },
    {
      type = 'simple',
      formatFunction = 'detailledTime',
      aggregateKey = 'bestTime',
      columnLabel = 'bigMap.progressLabels.bestTime',
      newBestKey = 'newBestTime',
    }
  }

  self.autoUiBigmap = {
    aggregates = {
      aggregatePrimary = {
        progressKey = 'default',
        aggregateKey = 'bestTime', -- select best time from detail progress key
        label = 'Best Time',
        formatFunction = 'detailledTime',
        type = 'simple'
      }
    }
  }
  self.bigMapIcon = {icon = 'map_mission_airace01'}
  self.missionTypeLabel = 'Extra AI Race'
end

function C:updateAttempt(attempt, newVersion) end

function C:processUserSettings(settings)
  self.userSettings = settings
end

function C:getUserSettingsData()
  local data = {}

  if self.missionTypeData.allowVehicleChoice then
    table.insert(data, {
      key = 'presetVehicleActive',
      label = 'missions.missions.general.userSettings.useProposedCar',
      type = 'bool',
      value = true
    })
  end

  if self.missionTypeData.lapCount >= 1 then -- assuming a circuit race
    table.insert(data, {
      key = 'lapCount',
      label = 'missions.missions.general.userSettings.laps',
      type = 'int',
      value = self.missionTypeData.lapCount
    })
  end

  if self.missionTypeData.allowReverseChoice then
    table.insert(data, {
      key = 'raceReversed',
      label = 'Reverse',
      type = 'bool',
      value = false
    })
  end

  table.insert(data, {
    key = 'racerCount',
    label = 'Racer Count',
    type = 'int',
    value = self.missionTypeData.racerCount,
    min = 1,
    max = 15
  })

  if self.missionTypeData.allowTraffic then
    table.insert(data, {
      key = 'trafficDifficulty',
      label = 'Traffic Difficulty',
      type = 'select',
      values = {
        {l = 'Off', v = 'Off'},
        {l = 'Normal', v = 'Normal'},
        {l = 'Hard', v = 'Hard'}
      },
      value = self.missionTypeData.trafficDifficulty
    })
  end

  table.insert(data, {
    key = 'rivalCarGroup',
    label = 'Rival Car Group',
    type = 'select',
    values = {
      {l = 'Normal', v = 'Normal'},
      {l = 'Tuned', v = 'Tuned'},
      {l = 'Race', v = 'Race'},
      -- {l = 'Rally', v = 'Rally'},
      -- {l = 'OffRoad', v = 'OffRoad'},
      {l = 'Derby', v = 'Derby'},
      {l = 'OneMake', v = 'OneMake'},
      {l = 'Extra', v = 'Extra'}
    },
    value = self.missionTypeData.rivalCarGroup
  })

  return data
end

function C:getRandomizedAttempt(progressKey)
  return gameplay_missions_progress.testHelper.randomAttemptType(), {
    medal = gameplay_missions_progress.testHelper.randomMedal(),
    place = round(gameplay_missions_progress.testHelper.randomNumber(1,4)),
    time = gameplay_missions_progress.testHelper.randomNumber(30,120)
  }
end

function C:getWorldPreviewRoute()
  if not self.cachedWorldPreviewRoute then
    -- load path from file
    local path = require('/lua/ge/extensions/gameplay/race/path')('New Path')
    path:onDeserialized(jsonReadFile(self.fgVariables.raceFile))
    path:autoConfig()
    local ret = {}
    for i, nId in ipairs(path.config.linearSegments) do
      local node = path.pathnodes.objects[nId]
      table.insert(ret, node.pos)
    end
    if path.config.closed then
      table.insert(ret, ret[1])
    end

    -- calculate in-world route
    local route = require('/lua/ge/extensions/gameplay/route/route')()
    route:setupPathMulti(ret)
    self.cachedWorldPreviewRoute = route.path
  end
  return self.cachedWorldPreviewRoute
end

return function(...) return gameplay_missions_missions.flowMission(C, ...) end
