-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local C = {}
C.__index = C

function C:init()
  ---- mandatory settings ----
  self:addDecoHeader("Mandatory Settings", "orange")

  self:addString("Start Screen Title", "startTitle", "For Example, 'My Mission Name'", 9000, {tooltip = "Title of the race.", isTranslation = true})

  self:addString("Start Screen Text", "startText", "For Example, 'Race against tough opponents!'", 9000, {tooltip = "Description of the race.", isTranslation = true})

  self:addString("Damage Failure Text", "damageFailText", "For Example, 'You got wrecked!'", 9000, {tooltip = "Text for when the player gets wrecked or stuck.", isTranslation = true})

  self:addDecoSeparator()

  self:addRace("Race Path File", "raceFile", "")
  self:addBool("Race Path Reverse", "raceReversed", false, nil, {tooltip = "If true, race path will be reversed."})
  self:addBool("Allow Reverse Race Choice", "allowReverseChoice", false, nil, {tooltip = "Should the player have the choice to use the race reversed?"})

  self:addDecoSeparator()

  ---- optional settings ----
  self:addDecoHeader("Optional Settings", "orange")

  self:addNumeric("Number of Laps", "lapCount", 1, {tooltip = "Values higher than one are only applicable for closed circuit layouts."})

  self:addDecoSeparator()

  self:addBool("Provided Vehicle", "presetVehicleActive", false, {
    self:addBool("Allow Player Vehicle Choice", "allowVehicleChoice", false, nil, {tooltip = "Should the player have the choice to use the preset vehicle in the mission settings?"})
  }, {boxText = "Should a preset vehicle be provided to the player?"})

  self:addDecoSeparator()

  self:addBool("Prefab", "prefabActive", false, {
    self:addFixedFile("Prefab File", "/deco.prefab")
  }, {boxText = "Should a prefab be used?", tooltip = "The prefab can contain decorations, triggers, or anything else."})

  self:addBool("Camera Path", "camPathActive", false, {
    self:addFixedFile("Camera Path File", "/intro.camPath.json")
  }, {boxText = "Should a custom camera path be used for the intro?", tooltip = "Cinematic camera view for the start screen."})

  self:addDecoSeparator()

  self:addBool("Use Groundmarkers", "useGroundmarkers", false, nil, {tooltip = "If true, decal markers will be drawn on the ground to show the race path."})

  self:addDecoSeparator()

  self:addBool("Fail Conditions", "damageFailActive", false, {
    self:addNumeric("Damage Limit", "damageLimit", 10000),
    self:addNumeric("Stopped Time Limit", "stoppedLimit", 0, {tooltip = "If past damage limit, the mission will fail if the player is stopped for this duration of time; set to 0 to fail instantly."})
  }, {boxText = "Should the mission fail if the player is damaged?"})

  self:addDecoSeparator()

  ---- racer settings ----
  self:addDecoHeader("AI Racer Setup", "orange")

  self:addNumeric("Default Amount", "racerCount", 1, {tooltip = "The flowgraph accepts a maximum of 16 AI racers.", min = 1, max = 16})

  self:addBool("Allow Rival Car Group Choice", "allowRivalCarGroupChoice", true, nil, {tooltip = "Should the player have the choice to use the rival car group?"})
  self:addDropdown("Rival Car Group","rivalCarGroup",{"Normal","Tuned","Race", "Rally", "OffRoad", "Derby", "Extra", "OneMake"}, "Tuned")
  self:addBool("Shuffle Group Order", "shuffleGroup", true, nil, {tooltip = "If true, the vehicle group order will be randomized."})

  self:addDecoSeparator()

  self:addNumeric("AI Aggression", "racerAggression", 1, {tooltip = "AI aggression value; higher values will make the AI take corners faster, but with a risk of losing control."})

  self:addNumeric("AI Skill Level", "racerSkill", 1, {tooltip = "AI skill level (from 0 to 1); higher values make the AI take more optimal racing lines."})

  self:addBool("AI Awareness", "racerAwareness", true, nil, {tooltip = "If true, vehicles will avoid collisions with each other."})

  self:addBool("Random AI Parameters", "randomizeAiParams", true, {
    self:addNumeric("Randomness Scale", "racerRandomnessScale", 1, {tooltip = "The strength of randomness for the AI parameters (from 0 to 2)"})
  }, {tooltip = "If true, AI vehicles will get some random parameters, so that their behavior will vary."})

  self:addBool("AI Rubberbanding", "racerRubberbandMode", false, nil, {tooltip = "If true, AI vehicles will lower their aggression while the player is falling behind."})

  self:addDecoSeparator()

  ---- traffic settings ----
  self:addDecoHeader("AI Traffic Setup", "orange")

  self:addBool("Allow Traffic", "allowTraffic", true, {
    self:addDropdown("Traffic Difficulty", "trafficDifficulty", {"Off", "Normal", "Hard"}, "Normal")
  }, {boxText = "Should traffic be allowed during the race?"})

  self:addDecoSeparator()

  ---- star settings ----
  self:addDecoHeader("Mission Goals", "orange")

  self:addNumeric("Laps Goal", "goalLaps", 3, {
    tooltip = "The number of laps needed to achieve the associated star.",
    associatedStars = {"finishGoalLaps"}
  })

  self:addNumeric("Time Goal", "goalTime", 180, {
    tooltip = "The time to beat to achieve the associated star.",
    associatedStars = {"beatGoalTime"}
  })

  self:addNumeric("Collisions Goal", "goalCollisions", 10, {
    tooltip = "The number of independent vehicle collisions needed to achieve the associated star.",
    associatedStars = {"dirtyRace"}
  })

  self:addNumeric("Total Damage", "totalDamage", 1000000, {
    tooltip = "The number of total damage needed to achieve the associated star.",
    associatedStars = {"totalDamage"}
  })

  self:addNumeric("Bounty", "bounty", 10000, {
    tooltip = "The number of independent bounty needed to achieve the associated star.",
    associatedStars = {"bounty"}
  })

  self:setAutoAdditionalAttributes("vehicle", true)
end

function C:customFunctionMoveTranslations(mission, missionEditor)
  -- get the mission instance, because it contains the activeStarCache.
  local instance = gameplay_missions_missions.getMissionById(mission.id)
  local defaultStarKeysSorted = deepcopy(instance.careerSetup._activeStarCache.defaultStarKeysSorted)

  for i = 1, #defaultStarKeysSorted do
    mission.careerSetup.starOutroTexts[defaultStarKeysSorted[i]] = mission.missionTypeData["completeText"]
  end
  mission.careerSetup.starOutroTexts["noStarUnlocked"] = mission.missionTypeData["passText"]

  -- set mission to dirty so we know it's been edited
  mission._dirty = true
end

return function(...) return gameplay_missions_missions.editorHelper(C, ...) end
