local M = {}
local logTag = 'mathkuro'

local DAMAGE_TBL_KEY = 10001
local DAMAGE_BY_VAL_TBL_KEY = 10002

local TOTAL_DAMAGE_BY_VAL_TBL_KEY = 20001
local TOTAL_DAMAGED_VEHICLE_NUM_TBL_KEY = 20002

-- ダメージ値を$ベースに換算する係数(要チューニング)
local DAMAGE_TO_VALUE = 25000

-- value未定義の車両のvalue値
local DEFAULT_VEHICLE_VALUE = 30000

-- ダメージ計算を行うダメージ[$]の閾値
local DAMAGE_CALC_THRESHOLD = 100

-- クラッシュカメラの設定
local SIM_SPEED = 0.25
local CRASH_CAMERA_MODE = 'external'


-- グローバル変数
-- クラッシュ系
local currentDamageTable = {}
local lastPlayerDamage = -1
local lastCameraName = ''
local isCrashCameraMode = false

local totalDamageValue = 0
local totalDamagedVehicleNum = 0

-- 被害総額算出対象とする車両距離
local DEFAULT_DISTANCE_THRESHOLD = 10
local distanceThreshold = DEFAULT_DISTANCE_THRESHOLD

-- 順位系
local currentRacePosition = 0
local startPositionTable = {}

-- 追跡情報系
local _lastOffensesTable = {}
local _lastHitCount = 0
local _isSpeeding = false
local _isRacing = false
local _isReckless = false
local _isWrongWay = false
local _isIntersection = false

local playerResetState = false
local function setPlayerResetState(state)
  playerResetState = state
  -- print('setCurrentRacePosition'..tostring(currentRacePosition))
end
local function getPlayerResetState()
  -- print('getCurrentRacePosition'..tostring(currentRacePosition))
  return playerResetState
end

local function setCurrentRacePosition(position)
  currentRacePosition = position
  -- print('setCurrentRacePosition'..tostring(currentRacePosition))
end
local function getCurrentRacePosition()
  -- print('getCurrentRacePosition'..tostring(currentRacePosition))
  return currentRacePosition
end

local lastOffense = ''
function onPursuitOffense(vid, tempData)
  lastOffense = tempData.key
  print(lastOffense)
end

local function getLastOffense()
  return lastOffense
end

local function getTable(key)
  if not currentDamageTable[key] then
    currentDamageTable[key] = {}
  end

  return currentDamageTable[key]
end

local function setDistanceThreshold(threshold)
  distanceThreshold = threshold
end

local function resetDamageValues(key)
  local _key = (type(key) == 'number') and key or DAMAGE_BY_VAL_TBL_KEY
  currentDamageTable[_key] = {}
  totalDamageValue = 0
  totalDamagedVehicleNum = 0
  return
end

local function getTotalDamageValue()
  return math.floor(totalDamageValue)
end

local function getDamagedVehicleNum()
  return totalDamagedVehicleNum
end

-- 車両毎のダメージテーブルを更新
local function updateDamagedVehicleTable(key)
  local _key = (type(key) == 'number') and key or DAMAGE_BY_VAL_TBL_KEY
  local _tbl = getTable(_key)
  local _totalDamageDiff = 0
  local _playerVeh = be:getPlayerVehicle(0)
  local _playerVid = _playerVeh:getId()
  local _playerPos = _playerVeh:getPosition()

  for vid, veh in activeVehiclesIterator() do
    local _currentDamage = 0
    -- 前回のダメージ値(初出の車両の場合は0)
    local _lastDamage = 0
    if _tbl[vid] then
      _lastDamage = _tbl[vid]['damage']
    end

    -- Value値が未設定の車両がたまにあるので補正
    local _value = core_vehicles.getVehicleDetails(vid).configs.Value
    if _value == nil then
      local _min_value = core_vehicles.getVehicleDetails(vid).model.aggregates.Value.min or (DEFAULT_VEHICLE_VALUE * (1 + math.random()))
      local _max_value = core_vehicles.getVehicleDetails(vid).model.aggregates.Value.max or (DEFAULT_VEHICLE_VALUE * (1 + math.random()))
      _value = (_min_value + _max_value)/2
    end

    if map.objects[vid] == nil then
      -- 'damage'が取得できない車両は諦めて0にしておく
      _currentDamage = 0
    else
      -- 車両価格の2倍まで許容？
      local _damageRate = (map.objects[vid]['damage'] / DAMAGE_TO_VALUE)
      _currentDamage = (2.0 > _damageRate) and (_damageRate * _value) or _value
    end

    if _currentDamage < _lastDamage or _currentDamage < DAMAGE_CALC_THRESHOLD then
      -- 以下の場合はテーブルから削除
      -- 前回値よりもダメージ量が小さい場合(車両リセット)
      -- DAMAGE_CALC_THRESHOLD以下(表示数削減)
      _tbl[vid] = nil
    else
      if vid ~= _playerVid then
        if (_tbl[vid] or veh:getPosition():distance(_playerPos) < distanceThreshold) then
          -- 登録済みまたはプレイヤー車両との距離が閾値未満の車両のみダメージ値更新

          if _tbl[vid] == nil then
            -- 新規登録車両は破壊台数に1加算
            totalDamagedVehicleNum = totalDamagedVehicleNum + 1
          end

          _tbl[vid] = {}
          -- 
          if string.find(core_vehicles.getVehicleDetails(vid).configs.Name, 'simple_traffic') == nil then
            _tbl[vid]['vname'] = core_vehicles.getVehicleDetails(vid).configs.Name
          else
            _tbl[vid]['vname'] = core_vehicles.getVehicleDetails(vid).model.Brand.." "..core_vehicles.getVehicleDetails(vid).model.Name
          end
          _tbl[vid]['damage'] = _currentDamage

          -- ダメージ差分
          _totalDamageDiff = _totalDamageDiff + (_currentDamage - _lastDamage)
        end
      end
    end
  end

  totalDamageValue = totalDamageValue + _totalDamageDiff
end

-- 車両毎のダメージテーブルを取得
local function getDamagedVehicleTable(key)
  local _key = (type(key) == 'number') and key or DAMAGE_BY_VAL_TBL_KEY
  return getTable(_key)
end

-- from lua\ge\extensions\gameplay\traffic\vehicle.lua C:checkOffenses()
local function checkOffenses() -- tests for vechicle offenses for police
  local _playerTrafficData = gameplay_traffic.getTrafficData()[be:getPlayerVehicleID(0)]
  if not _playerTrafficData then return end -- if not traffic activated

  local _speed = _playerTrafficData.speed
  local _tracking = _playerTrafficData.tracking
  local _collisions = _playerTrafficData.collisions

  local pursuit = _playerTrafficData.pursuit
  if not pursuit.mode or pursuit.mode < 1 then _lastHitCount = 0 return end -- if not pursuit started

  local minScore = clamp(_playerTrafficData.policeVars.strictness, 0, 0.8) -- offense threshold
  if _playerTrafficData.policeVars.strictness <= 0 then return end

  if _speed >= _tracking.speedLimit * 1.2 and not _isSpeeding then
    _lastOffensesTable['speeding'] = {value = _speed, maxLimit = _tracking.speedLimit, score = 100}
    _isSpeeding = true
  elseif  _speed < _tracking.speedLimit * 1.2 then
    _isSpeeding = false
  end
  if _speed >= _tracking.speedLimit * 2 and not _isRacing then
    _lastOffensesTable['racing'] = {value = _speed, maxLimit = _tracking.speedLimit, score = 200}
    _isRacing = true
  elseif _speed < _tracking.speedLimit * 2 then
    _isRacing = false
  end
  if _tracking.driveScore <= minScore and not _isReckless then
    _lastOffensesTable['reckless'] = {value = _tracking.driveScore, minLimit = minScore, score = 200}
    _isReckless = true
  elseif _tracking.driveScore > minScore then
    _isReckless = false
  end
  if _tracking.intersectionScore <= minScore and not _isIntersection then
    _lastOffensesTable['intersection'] = {value = _tracking.intersectionScore, minLimit = minScore, score = 200}
    _isIntersection = true
  elseif _tracking.intersectionScore > minScore then
    _isIntersection = false
  end
  if _tracking.directionScore <= minScore and not _isWrongWay then
    _lastOffensesTable['wrongWay'] = {value = _tracking.directionScore, minLimit = minScore, score = 150}
    _isWrongWay = true
  elseif _tracking.directionScore > minScore then
    _isWrongWay = false
  end

  if _lastHitCount < pursuit.hitCount then
    for id, coll in pairs(_collisions) do
      if coll.dot >= 0.2 then -- simple comparison to check if current vehicle is at fault for collision
        local veh = gameplay_traffic.getTrafficData()[id]

        _lastHitCount = _lastHitCount + 1
        if veh.role.name == 'police' and coll.inArea then -- always triggers if police was hit
          local score = veh.isPlayerControlled and 50 or 150
          _lastOffensesTable['hitPolice'] = {value = id, score = score}
        elseif pursuit.mode > 0 or coll.state == 'abandoned' then -- fleeing in a pursuit, or abandoning an accident
          _lastOffensesTable['hitTraffic'] = {value = id, score = 100}
        end
      end
    end
  end

end

local function getOffenses()
  local _offensesTable = _lastOffensesTable
  _lastOffensesTable = {}
  return _offensesTable
end

local function resetOffenses()
  _isSpeeding = false
  _isRacing = false
  _isWrongWay = false
end

-- プレイヤーのダメージ差分(前回呼び出し時との差分)を取得
local function getPlayerDamageDelta()
  local _playerDamageDelta = 0
  local _playerDamage = map.objects[be:getPlayerVehicle(0):getId()]["damage"]

  if _playerDamage < lastPlayerDamage then
    lastPlayerDamage = 0
  end

  if lastPlayerDamage == -1 then
    lastPlayerDamage = _playerDamage
    -- 初回呼び出し時に_playerDamageの値に関わらず差分は0として扱う
    _playerDamageDelta = 0
  else
    _playerDamageDelta = _playerDamage - lastPlayerDamage
    lastPlayerDamage = _playerDamage
  end

  return _playerDamageDelta
end

local function enableCrashCamera()
  -- 現在設定されているカメラを退避
  if isCrashCameraMode then
    -- クラッシュカメラ中はクラッシュカメラを起動しない
    log('D', logTag, 'crashCameraMode already enabled')
  else
    lastCameraName = core_camera.getActiveCamName()
    log('D', logTag, 'lastCameraName:'..lastCameraName)

    bullettime.set(SIM_SPEED)

    local _playerVelocity = be:getPlayerVehicle(0):getVelocity():length()
    log('D', logTag, '_playerVelocity:'.._playerVelocity)

    if _playerVelocity > 20 then
      core_camera.setByName(0, CRASH_CAMERA_MODE)
      core_camera.resetCamera(0)
    else
      core_camera.setByName(0, 'orbit')
      core_camera.cameraZoom(0.002)
      core_camera.rotate_yaw(0.1)
      local _vid = be:getPlayerVehicle(0):getId()
      -- core_camera.setFOV(_vid, 90)
    end

    isCrashCameraMode = true
  end
end

local function disableCrashCamera()
  if isCrashCameraMode then
    -- クラッシュカメラが有効化されている場合のみ解除処理を行う
    bullettime.set(1)
    core_camera.setByName(0, lastCameraName)
    core_camera.resetCamera(0)
    isCrashCameraMode = false
  else
    log('D', logTag, 'crashCameraMode is not enabled')
  end
end

local function _setStartPosition(vid)
  log('D', logTag, '_setStartPosition(' .. tostring(vid) .. ')')
  local veh = be:getObjectByID(vid)
  startPositionTable[vid] = {pos=vec3(veh:getPositionXYZ()), dirFront=veh:getDirectionVector(), dirUp=veh:getDirectionVectorUp()}
end

local function setStartPosition(vid)
  log('D', logTag, 'setStartPosition()')
  if not vid then
    for vid, _ in activeVehiclesIterator() do
      _setStartPosition(vid)
    end
  else
    _setStartPosition(vid)
  end
end

local function resetAllStartPosition()
  log('D', logTag, 'resetAllStartPosition()')
  for vid, startPos in pairs(startPositionTable) do
    log('D', logTag, 'vid=' .. tostring(vid))
    local veh = be:getObjectByID(vid)
    veh:queueLuaCommand('recovery.homePoint = ' .. serialize(startPos))
    veh:queueLuaCommand('recovery.loadHome()')
  end
end

local function resetAll()
  log('D', logTag, 'resetAll()')
  currentDamageTable = {}
  totalDamageValue = 0
  totalDamagedVehicleNum = 0
  lastPlayerDamage = -1
  distanceThreshold = DEFAULT_DISTANCE_THRESHOLD

  currentRacePosition = 0

  _lastOffensesTable = {}
  _lastHitCount = 0
end

M.setPlayerResetState = setPlayerResetState
M.getPlayerResetState = getPlayerResetState

M.resetAll = resetAll

M.setStartPosition = setStartPosition
M.resetAllStartPosition = resetAllStartPosition

M.getPlayerDamageDelta = getPlayerDamageDelta

M.enableCrashCamera = enableCrashCamera
M.disableCrashCamera = disableCrashCamera

M.checkOffenses = checkOffenses
M.getOffenses = getOffenses
M.resetOffenses = resetOffenses

M.setCurrentRacePosition = setCurrentRacePosition
M.getCurrentRacePosition = getCurrentRacePosition

M.onPursuitOffense = onPursuitOffense
M.getLastOffense = getLastOffense

M.resetDamageValues = resetDamageValues

M.setDistanceThreshold = setDistanceThreshold

M.getTotalDamageValue = getTotalDamageValue

M.getDamagedVehicleNum = getDamagedVehicleNum
M.updateDamagedVehicleTable = updateDamagedVehicleTable

M.getDamagedVehicleTable = getDamagedVehicleTable

return M
