angular
.module('beamng.apps')
.controller('AppCtrl', function ($scope, $mdSidenav) {
  // ---- 定数 ----
  const SETTINGS_PATH = '/ui/modules/apps/extraRaceMissionMathkuro/settings.json';

  // 言語設定
  const DEFAULT_LANG = 'en';
  const SUPPORT_LANGS = ['en', 'ja'];
  let currentLanguage = DEFAULT_LANG;

  let generalMessageTimerMap = new Map();

  // 近くの車両のダメージを画面表示する設定
  let damagedVehicleDisplayTimerMap = new Map();
  const DISPLAY_DAMAGE_THRESHOLD = 100;
  const DISPLAY_FRAMES = 120;

  // 被害情報
  let currentTotalDamage = 0;
  let damagedVehicleNum = 0;

  // レース情報
  let currentPosition = 0;
  let currentResetState = false;

  // 違反情報
  let totalPenalty = 0;
  let offensesDisplayTimerArray = new Array();
  let lastHeatLevel = 0;
  let currentHeatLevel = 0;
  const PURSUIT_RESULT_DISPLAY_MSEC = 8 * 1000;

  const GET_OFFENSES_INTERVAL = 1;
  let getOffensesDone = true;

  let crashCameraActivated = false;
  const CRASH_THRESHOLD = 10000;
  const CRASH_CAMERA_DURATION_SEC = 6;

  // 画面に表示する変数の初期化
  $scope.totalDamage = 0;
  $scope.totalDamageTitle = '';
  $scope.damagedVehicleNum = 0;
  $scope.damagedVehicleNumTitle = '';
  $scope.nearVehicleDamages = [];
  $scope.ja_jp = false;
  $scope.currentPosition = 0

  $scope.crashMessage = '';
  $scope.crashMessage2 = '';
  $scope.currentDamageString = '';

  $scope.generalMessage = '';

  $scope.pursuitResultMessage1 = '';
  $scope.pursuitResultMessage2 = '';
  $scope.pursuitResultMessage3 = '';

  // UIの表示設定・個別にON/OFF、ファイル保存
  $scope.uiSettings = {
    crashCamera: {
      title: 'Crash Camera',
      enabled: true
    },
    displayDamageInfo: {
      title: 'Damage Info',
      enabled: true
    },
    displayPursuitInfo: {
      title: 'Pursuit Info',
      enabled: true
    }
  }

  // 設定ファイル読み込み
  bngApi.engineLua('jsonReadFile(' + bngApi.serializeToLua(SETTINGS_PATH) + ')', (settings) => {
    if (settings != null) {
      console.log(settings);
      $scope.uiSettings = settings;
    }
  })


  // settings/Closeボタン押下時に実行される関数
  $scope.toggleLeft = buildToggler('left');
  $scope.closeLeft = closeToggler('left');
  // サイドメニューの開閉
  function buildToggler(componentId) {
    return function() {
      $mdSidenav(componentId).toggle();
    }
  }
  function closeToggler(componentId) {
    return function() {
      $mdSidenav(componentId).toggle();
      // サイドメニューを閉じた際に設定を保存しておく
      bngApi.engineLua('jsonWriteFile(' + bngApi.serializeToLua(SETTINGS_PATH) + ', ' + bngApi.serializeToLua($scope.uiSettings) + ', ' + bngApi.serializeToLua(true) + ')');
    }
  }


  // Get user language settings. ('<lang>_<country>')
  // ref: 0.25\lua\common\utils\languageMap.lua
  bngApi.engineLua('Lua:getSelectedLanguage()', (userLang) => {
    let _lang = userLang.split('_')[0];
    currentLanguage = (SUPPORT_LANGS.includes(_lang)) ? _lang : DEFAULT_LANG;
    if (currentLanguage == "ja") {
      $scope.ja_jp = true;
    }
    console.log('set language: ' + currentLanguage);
  });

  // for freeroam mode
  bngApi.engineLua('core_gamestate.state.state', (state) => {
    if (state == "freeroam") {
      bngApi.engineLua("extensions.unload('mathkuro_extraRaceMission')");
      bngApi.engineLua("extensions.loadAtRoot('lua/ge/extensions/mathkuro/extraRaceMission', 'mathkuro')");
      bngApi.engineLua('mathkuro_extraRaceMission.resetAll()');
    }
  });

  // フレーム毎のループ
  $scope.$on('streamsUpdate', function (event, data) {
    // 指定言語に合わせて表示に変更
    let locale = getLocale(currentLanguage);

    if ($scope.uiSettings.displayDamageInfo.enabled) {
      // ダメージ情報更新
      bngApi.engineLua('mathkuro_extraRaceMission.updateDamagedVehicleTable()');
      // 違反情報更新
      bngApi.engineLua('mathkuro_extraRaceMission.checkOffenses()');

      // 被害総額取得
      bngApi.engineLua('mathkuro_extraRaceMission.getDamagedVehicleNum()', (_num) => {
        damagedVehicleNum = _num;
      });

      // 被害総額取得
      bngApi.engineLua('mathkuro_extraRaceMission.getTotalDamageValue()', (damage) => {
        currentTotalDamage = damage;
      });
    }

    // 順位取得
    bngApi.engineLua('mathkuro_extraRaceMission.getCurrentRacePosition()', (position) => {
      currentPosition = position;
    });

    // リセット状態取得
    bngApi.engineLua('mathkuro_extraRaceMission.getPlayerResetState()', (resetState) => {
      currentResetState = resetState;
      // console.log(currentResetState);
    });

    if ($scope.uiSettings.displayDamageInfo.enabled) {
      bngApi.engineLua('extensions.gameplay_police.getPursuitData()', (pursuitData) => {
        if (pursuitData) {
          currentHeatLevel = pursuitData['mode'];

          if (currentHeatLevel > lastHeatLevel) {
            generalMessageTimerMap.set("heatLevel", {remain: 150, message: locale.getHeatLevelText(currentHeatLevel)});
            lastHeatLevel = currentHeatLevel;
          }

        }
      });

      // 車両毎の被害額取得
      bngApi.engineLua('mathkuro_extraRaceMission.getDamagedVehicleTable()', (_damagedVehicleTable) => {
        Object.keys(_damagedVehicleTable).forEach((key) => {
          let _vname = _damagedVehicleTable[key]['vname'];
          let _damage = _damagedVehicleTable[key]['damage'];
          let _last_val = damagedVehicleDisplayTimerMap.get(key);
          if (typeof _last_val === 'undefined') {
            // 表示数が増えて見にくくなってしまうため、ダメージ値が一定以上の被害車両のみ追加
            if (_damage >= DISPLAY_DAMAGE_THRESHOLD) {
              damagedVehicleDisplayTimerMap.set(key, {vname: _vname, damage: _damage, remain: DISPLAY_FRAMES});
            }
          } else {
            // UPDATE
            if (_last_val['remain'] <= 0 && _damage < _last_val['damage']) {
              damagedVehicleDisplayTimerMap.delete(key)
            } else {
              _last_val['damage'] = _damage;
              damagedVehicleDisplayTimerMap.set(key, _last_val);
            }
          }
        });
      });
    }

    if ($scope.uiSettings.displayPursuitInfo) {
      if (getOffensesDone) {
        // console.log('setTimeout: getOffenses')
        getOffensesDone = false;
        setTimeout(
          function() {
            // 直近の違反取得
            bngApi.engineLua('mathkuro_extraRaceMission.getOffenses()', (_offensesTable) => {
              if (_offensesTable) {
                Object.keys(_offensesTable).forEach((key) => {
                  offensesDisplayTimerArray.push({offense: key, remain: DISPLAY_FRAMES})
                  totalPenalty = totalPenalty + locale.offenses[key].penalty;
                  console.log('key: ' + key);
                })
              }
            });
            getOffensesDone = true;
          }, (GET_OFFENSES_INTERVAL * 1000)
        );
      }
    }

    // クラッシュカメラ
    if ($scope.uiSettings.crashCamera.enabled) {
      enableCrashCamera();
    }

    // 画面表示更新
    $scope.$evalAsync(function() {
      updateCrashMessageText();

      let _damages = [];
      damagedVehicleDisplayTimerMap.forEach((val, key) => {
        val['remain'] = val['remain'] - 1;
        if (val['remain'] > 0) {
          // 表示期間が残っている場合のみ画面表示
          // console.log('name: ' + val['vname'] + ', damage:' + val['damage'] + ', remain:' + val['remain']);
          _damages.push({name: val['vname'], damage: locale.valueToLocalString(val['damage'])});
        }
      });

      let _offenses = [];
      offensesDisplayTimerArray.forEach((val) => {
        val['remain'] = val['remain'] - 1;
        if (val['remain'] > 0) {
          // 表示期間が残っている場合のみ画面表示
          // console.log('key: ' + key + ', remain:' + val['remain']);
          _offenses.push({offense: locale.offenses[val['offense']].name, penalty: locale.getPenaltyText(locale.offenses[val['offense']].penalty)});
        }
      });

      if (currentHeatLevel < lastHeatLevel) {
        setTimeout(
          function() {
            $scope.pursuitResultMessage1 = '';
            $scope.pursuitResultMessage2 = '';
            $scope.pursuitResultMessage3 = '';
            totalPenalty = 0;
          }, (PURSUIT_RESULT_DISPLAY_MSEC)
        );

        if (-1 == currentHeatLevel) {
          $scope.pursuitResultMessage1 = locale.getPursuitResultText(true);
        } else {
          $scope.pursuitResultMessage1 = locale.getPursuitResultText(false);
        }
        $scope.pursuitResultMessage2 = locale.getTotalpenaltyText(totalPenalty);
        $scope.pursuitResultMessage3 = locale.getFinedText(totalPenalty);

        lastHeatLevel = 0;
      }

      if (currentResetState) {
        $scope.generalMessage = locale.playerResetMessage;
        crashCameraActivated = false;
      } else if (generalMessageTimerMap.size > 0) {
        let key = generalMessageTimerMap.keys().next().value;
        let _remain = generalMessageTimerMap.get(key)['remain'];
        if (_remain > 0) {
          generalMessageTimerMap.get(key)['remain'] = _remain - 1;
          $scope.generalMessage = generalMessageTimerMap.get(key)['message'];
        } else {
          generalMessageTimerMap.delete(key);
          $scope.generalMessage = '';
        }
      } else {
        $scope.generalMessage = '';
      }

      switch (currentPosition) {
        case 0:
          $scope.currentPosition = '';break;
        case 1:
          $scope.currentPosition = '' + currentPosition + 'st';break;
        case 2:
          $scope.currentPosition = '' + currentPosition + 'nd';break;
        case 3:
          $scope.currentPosition = '' + currentPosition + 'rd';break;
        default:
          $scope.currentPosition = '' + currentPosition + 'th';
      }

      $scope.totalDamageTitle = locale.totalDamageTitle;
      $scope.totalDamage = locale.valueToLocalString(currentTotalDamage);
      $scope.damagedVehicleNumTitle = locale.damagedVehicleNumTitle;
      $scope.damagedVehicleNum = damagedVehicleNum
      $scope.nearVehicleDamages = _damages;

      $scope.offensesTable = _offenses;

      if (totalPenalty < 0) {
        $scope.totalPenaltyText = locale.getTotalpenaltyText(totalPenalty);
      } else {
        $scope.totalPenaltyText = "";
      }
    });
  });

  // -- クラッシュカメラの処理 --
  function enableCrashCamera() {
    bngApi.engineLua('mathkuro_extraRaceMission.getPlayerDamageDelta()', (_playerDamageDelta) => {
      // プレイヤー車両の1回のダメージ値が閾値を超えるとクラッシュカメラ起動

      if (_playerDamageDelta > CRASH_THRESHOLD) {
        // クラッシュカメラ有効化
        bngApi.engineLua('mathkuro_extraRaceMission.enableCrashCamera()');
        crashCameraActivated = true;

        // 一定時間でクラッシュカメラ無効化
        // 実行順の制御はUI側に持たせたいのでタイムアウト処理は画面側で行う
        setTimeout(
          function() {
            bngApi.engineLua('mathkuro_extraRaceMission.disableCrashCamera()');
            crashCameraActivated = false;
          }, (CRASH_CAMERA_DURATION_SEC * 1000)
        );
      }
    });
  }

  // -- クラッシュ時に表示する文字を更新 --
  function updateCrashMessageText() {
    if (!crashCameraActivated || !$scope.uiSettings.displayDamageInfo.enabled) {
      // クラッシュ中(クラッシュカメラ起動期間中)以外は何も表示しない
      $scope.crashMessage = '';
      $scope.crashMessage2 = '';
      $scope.currentDamageString = '';
      return;
    }

    let locale = getLocale(currentLanguage);
    // console.log(locale.totalDamageTitle + '  ' + locale.valueToLocalString(currentTotalDamage));

    $scope.crashMessage = locale.crashMessage;
    $scope.crashMessage2 = locale.crashMessage2;
    $scope.currentDamageString = locale.totalDamageTitle + '  ' + locale.valueToLocalString(currentTotalDamage);
  }

  // ---- UI上に表示する言語の設定 ----
  function getLocale(localeName){
    if (localeName == 'ja') {
      return new JaLocale();
    } else {
      return new EnLocale();
    }
  };

  class LocalInfoBase {
    totalDamageTitle = 'CRASH TOTAL';
    damagedVehicleNumTitle = 'WRECKS';

    crashMessage = 'CRASHED! ACCIDENT!!';
    crashMessage2 = 'RECKLESS!!';

    bustedMessage = 'YOU GOT BUSTED!!'

    playerResetMessage = 'Reset to last checkpoint. (Press reset again if you want to restart from start position.)'

    offenses = {
      speeding: {name: 'Speeding', penalty: -1},
      racing: {name: 'Racing', penalty: -12},
      hitPolice: {name: 'Hit(Police)', penalty: -35},
      hitTraffic: {name: 'Hit(Traffic)', penalty: -5},
      reckless: {name: 'Reckless', penalty: -35},
      intersection: {name: 'Intersection', penalty: -2},
      wrongWay: {name: 'Wrong Way', penalty: -2}
    };

    valueToLocalString(damage) {
      return '$' + damage.toLocaleString(undefined, { maximumFractionDigits: 0 });
    }
    getPenaltyText(penalty) {
      return String(penalty) + ' Points';
    }
    getTotalpenaltyText(penalty) {
      return 'Total Penalty Points: ' + String(penalty);
    }
    getPursuitResultText(isBusted) {
      if (isBusted) {
        return 'You got busted.';
      }
      return 'Escaped.';
    }
    getFinedText(penalty) {
      return 'Fined: $' + Math.abs(penalty) * 100;
    }
    getHeatLevelText(currentHeatLevel) {
      return 'Heat Level increased: Level ' + currentHeatLevel;
    }
  };

  class EnLocale extends LocalInfoBase {};

  class JaLocale extends LocalInfoBase {
    totalDamageTitle = '被害総額';
    damagedVehicleNumTitle = '被害台数';

    crashMessage = '重大事故発生';
    crashMessage2 = '危険運転';

    bustedMessage = '検挙されました'

    playerResetMessage = '前回のチェックポイントにリスポーンしました。(もう一度リセットを押すとレースをリスタートします)'

    offenses = {
      speeding: {name: '速度超過', penalty: -1},
      racing: {name: '速度超過[重度]', penalty: -12},
      hitPolice: {name: '緊急車両妨害', penalty: -35},
      hitTraffic: {name: '当て逃げ', penalty: -5},
      reckless: {name: '危険運転', penalty: -35},
      intersection: {name: '信号無視', penalty: -2},
      wrongWay: {name: '逆走', penalty: -2}
    };

    // 京の桁まで対応
    maxDisplayDigits = 10000 * 10000 * 10000 * 10000 * 10000;

    valueToLocalString(damage) {
      let _damage = damage * 120;

      if (_damage > this.maxDisplayDigits) {
        return '￥' + (damage * 120).toLocaleString(undefined, { maximumFractionDigits: 0 });
      } else {
        let _str = String(Math.round(_damage));
        let keta = ['', '万', '億', '兆', '京'];
        let nums = _str.replace(/(\d)(?=(\d\d\d\d)+$)/g, '$1,').split(',').reverse();
        let data = '';
        for (let i = 0; i < nums.length; i++) {
          if ((nums.length - i) > 2) {
            continue;
          }

          if (!nums[i].match(/^[0]+$/)) {
            data = nums[i].replace(/^[0]+/g, '') + keta[i] + data;
          }
        }
        if (data == '') {
          data = '0';
        }
        return data + '円';
      }
    }

    getPenaltyText(penalty) {
      return String(penalty) + '点';
    }
    getTotalpenaltyText(penalty) {
      return '違反点数: ' + String(penalty) + '点';
    }
    getPursuitResultText(isBusted) {
      if (isBusted) {
        return '検挙されました。';
      }
      return '逃走成功';
    }
    getFinedText(penalty) {
      let message = '';
      let _penalty = Math.abs(penalty)
      // 前科持ちについて考えるのはめんどいわ。
      if (_penalty < 6) {
        message = message + '';
      }else if (_penalty <= 8) {
        message = message + '免許停止 30日間です。'
      } else if (_penalty <= 11) {
        message = message + '免許停止 60日間です。'
      } else if (_penalty <= 14) {
        message = message + '免許停止 90日間です。'
      } else {
        message = message + '免許取消 ' + Math.floor(_penalty / 15) + '年です。'
      }
      return message;
    }
    getHeatLevelText(currentHeatLevel) {
      return 'ヒートレベルが ' + currentHeatLevel + ' に引き上げられました。';
    }
  };

})
.directive('extraRaceMission', [function () {
  return {
    templateUrl: '/ui/modules/apps/extraRaceMissionMathkuro/app.html',
    replace: true,
    restrict: 'EA',
    link: function (scope, element, attrs) {
      'use strict';

        // let obj_main = angular.element(element[0].children[2]);
        let obj_main = angular.element(document.getElementById('fugitive-dash'));
        obj_main.on('load', function () {
          if (scope.uiSettings.displayPursuitInfo.enabled) {
            let svg = obj_main[0].contentDocument;
            scope.$on('streamsUpdate', function (event, data) {
              bngApi.engineLua('extensions.gameplay_police.getPursuitData()', (pursuitData) => {
                if (pursuitData) {parsePursuitData(svg, pursuitData);}
              });
            });
          }
        });

      // 秒単位のsec(int型)をHH:MM:SS.fff形式に変換する。digitsは小数点以下の桁数。
      function sec2TimeStr(sec, digits = 0) {
        if (!Number.isFinite(sec)) {return '00:00:00';}
        var secStr = ('00' + (sec % 60).toFixed(digits)).slice(-(3 + digits));
        var minStr = ('00' + (Math.floor(sec / 60) % 60)).slice(-2) + ':';
        // hourは1時間超えた時以外は非表示（割と幅とって邪魔だから）
        var hourStr = (Math.floor(sec / 3600) > 0) ? ('00' + Math.floor(sec / 3600)).slice(-2) + ':' : '';
        return hourStr + minStr + secStr;
      }

      // ---- 追跡情報の解析 ----
      function parsePursuitData(svg, pursuitData) {
        if (!('mode' in pursuitData) || !('timers' in pursuitData)) {return;}
        let heatLevel = pursuitData['mode'];
        let timers = pursuitData['timers'];
        let pursuitTime = (timers['main'] > 0) ? timers['main'] : 0;
        let arrestLimit = timers['arrestValue'] * 100;
        let evadeLimit = timers['evadeValue'] * 100;
        if (evadeLimit > arrestLimit) {
          arrestLimit = 0;
        } else {
          evadeLimit = 0;
        }

        // 逮捕判定の間は時刻等の数値が想定外の値を示すため描画を更新しない(各値の初期化タイミングが異なる為？)
        if (heatLevel > -1) {
          // let svg = angular.element(document.getElementById('fugitive-dash')).contentDocument;
          svg.getElementById('arrest-limit').setAttribute("width", arrestLimit);
          svg.getElementById('evade-limit').setAttribute("width", evadeLimit);

          // リアルタイムに更新する値の部分はtoFixedで桁数を固定しておかないと見栄えがイマイチ。表示幅的にカンマもない方が良い。
          svg.getElementById('pursuitTime').textContent = sec2TimeStr(pursuitTime, 2);
          svg.getElementById('hitCount').textContent = pursuitData['hitCount'].toLocaleString();
          svg.getElementById('bounty').textContent = pursuitData['score'].toFixed(1);
        }
      };

    }
  }
}])
